/*
* Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
* Copyright (C) 2015-2016 Mellanox Technologies Ltd. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:

* 1. Redistributions of source code must retain the above copyright
* notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
* notice, this list of conditions and the following disclaimer in the
* documentation and/or other materials provided with the distribution.
* 3. Neither the name of the copyright holder nor the names of its
* contributors may be used to endorse or promote products derived from
* this software without specific prior written permission.

* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
* HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "api/clx_api.h"

#define MAX_SEVERITY_LEN 2
#define MAX_FACILITY_LEN 3
#define MAX_HOSTNAME_LEN 255
#define MAX_APP_NAME_LEN 48
#define MAX_PROCID_LEN 128
#define MAX_MSGID_LEN 32
#define MAX_STRUCTURED_DATA_LEN 512
#define MAX_MSG_LEN 1024
#define MAX_SYSLOG_MESSAGE_LEN 2048


typedef struct test_event_provider_info_t {
    uint8_t ev_syslog_type_index;
    // add implementation-specific fields here
} test_event_provider_info_t;


/**
 * This struct is designed to meet Syslog's strict formatting requirements,
 * ensuring compatibility with its standard field structure. Unlike other examples,
 * this serves as a specific demonstration of clxapi usage for Syslog integration.
 */
typedef struct __attribute__((packed)) test_syslog_event_t {
    char                severity[MAX_SEVERITY_LEN];
    char                facility[MAX_FACILITY_LEN];
    char                timestamp[25];
    char                hostname[MAX_HOSTNAME_LEN];
    char                app_name[MAX_APP_NAME_LEN];
    uint8_t             procid;
    char                msgid[MAX_MSGID_LEN];
    char                structured_data[MAX_STRUCTURED_DATA_LEN];
    char                msg[MAX_MSG_LEN];
} test_syslog_event_t;


bool test_event_provider_initialize(clx_api_context_t* ctx, clx_api_provider_t* provider) {
    test_event_provider_info_t* info = (test_event_provider_info_t*) provider->data;
    //
    // The info below describes data event entries,
    // and must exactly describe the structs defined above.
    //

    clx_api_event_field_info_t fields_syslog_event[] = {
        { "severity",   "Severity",               "char",     CLX_FIELD_VALUE_ABSOLUTE, MAX_SEVERITY_LEN },
        { "facility",   "Facility",               "char",     CLX_FIELD_VALUE_ABSOLUTE, MAX_FACILITY_LEN },
        { "time",       "Event timestamp",        "char",     CLX_FIELD_VALUE_ABSOLUTE, 25 },
        { "host",       "Hostname",               "char",     CLX_FIELD_VALUE_ABSOLUTE, MAX_HOSTNAME_LEN },
        { "ident",      "Application Name",       "char",     CLX_FIELD_VALUE_ABSOLUTE, MAX_APP_NAME_LEN },
        { "pid",        "Process ID",             "uint8_t",  CLX_FIELD_VALUE_ABSOLUTE, 1 },
        { "msgid",      "Message ID",             "char",     CLX_FIELD_VALUE_ABSOLUTE, MAX_MSGID_LEN },
        { "extradata",  "Structured Data",        "char",     CLX_FIELD_VALUE_ABSOLUTE, MAX_STRUCTURED_DATA_LEN },
        { "message",    "Message Content",        "char",     CLX_FIELD_VALUE_ABSOLUTE, MAX_MSG_LEN },
    };

    if (CLX_API_OK != clx_api_add_event_type(ctx, "syslog_event", fields_syslog_event,
                                            NUM_OF_FIELDS(fields_syslog_event), &info->ev_syslog_type_index)) {
        return false;
    }

    return true;
}

static void convert_timestamp_to_iso(uint64_t timestamp, char* buffer, size_t buffer_size) {
    uint64_t seconds = timestamp / 1000000;
    uint64_t microseconds = timestamp % 1000000;
    struct tm gm_time;

    time_t raw_time = (time_t)seconds;

    if (gmtime_r(&raw_time, &gm_time) == NULL) {
        snprintf(buffer, buffer_size, "Invalid timestamp");
        return;
    }

    snprintf(buffer, buffer_size, "%04d-%02d-%02dT%02d:%02d:%02d.%06luZ",
             gm_time.tm_year + 1900, gm_time.tm_mon + 1, gm_time.tm_mday,
             gm_time.tm_hour, gm_time.tm_min, gm_time.tm_sec, microseconds);
}


clx_api_provider_t* get_provider(void) {
    clx_api_provider_t* provider = calloc(1, sizeof(clx_api_provider_t));
    clx_api_version_t ver = { { 1, 0, 0 } };  // compiler bug.
    provider->name = strdup("My_provider");
    provider->version = ver;
    provider->initialize = &test_event_provider_initialize;

    // for storing the event type idx returned for each event.
    // can be used for additional provider-specific info
    provider->data = calloc(1, sizeof(test_event_provider_info_t));
    return provider;
}

void destroy_provider(clx_api_provider_t* provider) {
    free(provider->name);
    free(provider->data);
    free(provider);
}

#define MULTISOURCE
// enable MULTISOURCE to clone api context to use it for the second source

// #define WITH_IPC
// enable ipc transport
// IPC transport sents the data to collector 'clx' process from this app

int main(void) {
    int k;
    int result = 0;
    char iso8601[25];

    clx_api_params_t params = {0};
    params.max_file_size = 1 * 1024 * 1024;  // 1 MB
    params.max_file_age = 60 * 60 * 1000000L;  // 1 hour
    params.buffer_size = 0;  // use default

    params.schema_path = "tmp/example/schema";
    params.data_root = "tmp/example";

    params.source_id = "S1";
    params.source_tag = "example_event";
    params.file_write_enabled = true;

    params.ipc_enabled = 0;

#ifdef WITH_IPC
    params.ipc_enabled     = 1;
    params.ipc_sockets_dir = strdup("/tmp/ipc_sockets");  // should be the same ipc_sockets_dir from clx_config.ini

    params.ipc_max_reattach_time_msec = 5000;  // 5 seconds for overal reattach procedure
    params.ipc_max_reattach_tries     = 10;    // 10 tries during reattach procedure

    params.ipc_socket_timeout_msec    = 3000;  // timeout for UD socket
#endif


    // from common/defs.h:
    // #define CLX_DATA_PATH_TEMPLATE          "{{year}}/{{month}}{{day}}/{{source}}/{{tag}}{{id}}.bin"
    // params.data_path_template = strdup(CLX_DATA_PATH_TEMPLATE); // was in ipc branch
    params.data_path_template = CLX_API_DATA_PATH_TEMPLATE;

    // params.fb_enable = true;

    clx_api_provider_t* provider = get_provider();
    if (provider == NULL) {
        printf("Failed to initialize provider\n");
        return -1;
    }

    clx_api_context_t* ctx = clx_api_create_context(&params, provider);


#ifdef MULTISOURCE
    // For testing the 'clx_api_clone_context()' function
    clx_api_params_t params_2 = params;  // copy element by element
    params_2.source_id = "S2";
    params_2.source_tag = "example_event";
    params_2.schema_path = "tmp/example/schema";
    params_2.data_root = "tmp/example";
    params_2.data_path_template = CLX_API_DATA_PATH_TEMPLATE;
    // params_2.fb_enable = true;

    clx_api_context_t* ctx2 = clx_api_clone_context(ctx, &params_2);
#endif


    test_event_provider_info_t* info = (test_event_provider_info_t*) provider->data;


    test_syslog_event_t syslog_event = {{0}};

    strncpy(syslog_event.severity, "1", MAX_SEVERITY_LEN);
    strncpy(syslog_event.facility, "11", MAX_FACILITY_LEN);
    strncpy(syslog_event.hostname, "my-host", MAX_HOSTNAME_LEN);
    strncpy(syslog_event.app_name, "my-app", MAX_APP_NAME_LEN);
    strncpy(syslog_event.structured_data, "[exampleSDID@32473 iut=\"3\" eventSource=\"Application\"]",
            MAX_STRUCTURED_DATA_LEN);
    strncpy(syslog_event.msg, "Application event message", MAX_MSG_LEN);

#ifdef MULTISOURCE
    test_syslog_event_t syslog_event2 = {{0}};

    strncpy(syslog_event2.severity, "1", MAX_SEVERITY_LEN);
    strncpy(syslog_event2.facility, "11", MAX_FACILITY_LEN);
    strncpy(syslog_event2.hostname, "my-host2", MAX_HOSTNAME_LEN);
    strncpy(syslog_event2.app_name, "my-app2", MAX_APP_NAME_LEN);
    strncpy(syslog_event2.structured_data, "[exampleSDID@32473 iut=\"3\" eventSource=\"Application\"]",
            MAX_STRUCTURED_DATA_LEN);
    strncpy(syslog_event2.msg, "Application event2 message", MAX_MSG_LEN);
#endif

    for (k = 0; k < 1000; k++) {
            clx_api_timestamp_t ts = clx_api_get_timestamp();

            convert_timestamp_to_iso(ts, iso8601, sizeof(iso8601));
            strncpy(syslog_event.timestamp, iso8601, sizeof(syslog_event.timestamp));
            syslog_event.procid = k + 1;
            snprintf(syslog_event.msgid, MAX_MSGID_LEN, "ID%d", k + 1);
            bool ok = clx_api_write_event(ctx, &syslog_event, info->ev_syslog_type_index, 1);
            if (!ok) {
                printf("Failed to write ev4_syslog_type_index event %d\n", k);
                result = 1;
                break;
            }
#ifdef MULTISOURCE
            syslog_event2.procid = k + 1;
            strncpy(syslog_event2.timestamp, iso8601, sizeof(syslog_event.timestamp));
            snprintf(syslog_event2.msgid, MAX_MSGID_LEN, "ID%d", k + 1);
            ok = clx_api_write_event(ctx2, &syslog_event2, info->ev_syslog_type_index, 1);
            if (!ok) {
                printf("Failed to write ev4_syslog_type_index event %d\n", k);
                result = 1;
                break;
            }
#endif
    }

    destroy_provider(provider);
#ifdef MULTISOURCE
    clx_api_destroy_context(ctx2);
#endif
    clx_api_destroy_context(ctx);
    return result;
}
